/*
Copyright 2015 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package authtls

import (
	"github.com/pkg/errors"
	"k8s.io/kubernetes/pkg/apis/extensions"

	"k8s.io/ingress/core/pkg/ingress/annotations/parser"
	ing_errors "k8s.io/ingress/core/pkg/ingress/errors"
	"k8s.io/ingress/core/pkg/ingress/resolver"
	"k8s.io/ingress/core/pkg/k8s"
)

const (
	// name of the secret
	authTLSSecret = "ingress.kubernetes.io/auth-tls-secret"
)

type authTLS struct {
	certResolver resolver.AuthCertificate
}

// NewParser creates a new TLS authentication annotation parser
func NewParser(resolver resolver.AuthCertificate) parser.IngressAnnotation {
	return authTLS{resolver}
}

// ParseAnnotations parses the annotations contained in the ingress
// rule used to use an external URL as source for authentication
func (a authTLS) Parse(ing *extensions.Ingress) (interface{}, error) {
	str, err := parser.GetStringAnnotation(authTLSSecret, ing)
	if err != nil {
		return nil, err
	}

	if str == "" {
		return nil, ing_errors.NewLocationDenied("an empty string is not a valid secret name")
	}

	_, _, err = k8s.ParseNameNS(str)
	if err != nil {
		return nil, ing_errors.NewLocationDenied("an empty string is not a valid secret name")
	}

	authCert, err := a.certResolver.GetAuthCertificate(str)
	if err != nil {
		return nil, ing_errors.LocationDenied{
			Reason: errors.Wrap(err, "error obtaining certificate"),
		}
	}

	return authCert, nil
}
